/* system includes */
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <fcntl.h>
#include <sys/types.h>
#include <sys/un.h>

/* openssl includes */
#include <openssl/des.h>

/* heimdal includes */
#include <krb5.h>

/* cracklib includes */
#include <packer.h>

/* our own ... */
#include "pw_check.h"

const int version = 0; /* exported, kpasswdd checks for this! */

/* check password for character categories, wrong characters */
char * check_characters(char *);

/* check principal, password against (modified) cracklib */
char * check_cracklib(krb5_context, char *, char *);

/* check for password reuse */
char * check_pw_reuse(krb5_context, char *, char *);

/* save the new password in password history database */
char * save_pw_history(krb5_context, char *, char *);

/* write user name to a unix socket to notify other processes about the
   password change */
char * notify_socket(krb5_context, char *);

/* exported functions */

/* for test purposes like the test program to check if a password is
   OK for use */
char * test_passwd(krb5_context context,
		   krb5_principal principal,
		   krb5_data *password)
{
	char *val	= NULL,
	     *user	= principal->name.name_string.val[0], /* XXX */
	     *pw	= malloc(password->length + 1),
	     tmpstr[MESSAGE_LEN];

	if ( pw == NULL ) return "Out of memory";

	memcpy(pw, password->data, password->length);
	pw[password->length] = '\0';

	val = check_characters(pw);
	if ( val == NULL ) val = check_cracklib(context, user, pw);

	if ( val != NULL )
	{
		sprintf(tmpstr, "Password NOT ok! Reason: %s", val);
		strcpy(val, tmpstr);
	}

	/* overwrite password */
	memset(pw, 0, password->length);
	free(pw);

	return val;
}

/* the actual check function -> should be called by kpasswdd */
char * passwd_check(krb5_context context,
		    krb5_principal principal,
		    krb5_data *password)
{
	char *val	= NULL,
	     *user	= principal->name.name_string.val[0], /* XXX */
	     *pw	= malloc(password->length + 1),
	     tmpstr[MESSAGE_LEN];

	if ( pw == NULL ) return "Out of memory";

	memcpy(pw, password->data, password->length);
	pw[password->length] = '\0';

	val = check_characters(pw);
	if ( val == NULL ) val = check_cracklib(context, user, pw);
	if ( val == NULL ) val = check_pw_reuse(context, user, pw);
	if ( val == NULL ) val = save_pw_history(context, user, pw);
	if ( val == NULL ) val = notify_socket(context, user);

	if ( val != NULL )
	{
		sprintf(tmpstr, "Password NOT changed! Reason: %s", val);
		strcpy(val, tmpstr);
	}

	/* overwrite password */
	memset(pw, 0, password->length);
	free(pw);

	return val;
}

/* internal functions */

char * check_characters(char * password)
{
	const char *bad_characters = "\\|",
		   *special_characters = "#$%&'()*+,-./:;<=>?@[]^_`{}~";
	static char val[MESSAGE_LEN];
	int	i, j, categories = 0;

	val[0] = '\0';
	/* check for bad characters */
	for ( i = 0; i < strlen(password); i++ )
		for ( j = 0; j < strlen(bad_characters); j++ )
			if ( password[i] == bad_characters[j] )
			{
				sprintf(val, "do not use the following "
					"characters: %s",
					bad_characters);
				return val;
			}	

	for ( i = 'a'; i <= 'z' ; i++ )
		if ( strchr(password, i) != NULL )
		{
			categories++;
			break;
		}

	for ( i = 'A'; i <= 'Z' ; i++ )
		if ( strchr(password, i) != NULL )
		{
			categories++;
			break;
		}

	for ( i = '0'; i <= '9' ; i++ )
		if ( strchr(password, i) != NULL )
		{
			categories++;
			break;
		}

	for ( i = 0; i < strlen(special_characters); i++ )
		if ( strchr(password, special_characters[i]) != NULL )
		{
			categories++;
			break;
		}

	if ( categories < 3 )
		strcpy(val, "the password must contain at least 3 different "
			"character groups");

	return strlen(val) > 0 ? val : NULL;
}

char * check_cracklib(krb5_context context, char * user, char * password)
{
	const char *cracklib = krb5_config_get_string(context, NULL,
			"password_quality", "cracklib_db", NULL);
	static char val[MESSAGE_LEN];
	char	*clret = NULL;

	if ( cracklib == NULL )
	{
		strcpy(val, "no cracklib database given");
		return val;
	}

	val[0] = '\0';
	/* do the actual check */
	if ( (clret = FascistCheck(password, cracklib, user)) != NULL )
		strcpy(val, clret);
	return strlen(val) > 0 ? val : NULL;
}

int check_pw_equal(char * crypted, char * password)
{
	char *crypt_pw, salt[3];

	/* the first two characters of the crypted password are the salt */
	strncpy(salt, crypted, 2);
	/* DES crypt the new password with this salt */
	crypt_pw = crypt(password, salt);

	return strcmp(crypt_pw, crypted);
}

char * check_pw_reuse(krb5_context context, char * user, char * password)
{
	char	account[32], crypted[15], history[MAX_HIST * 14 + 1];
	static char val[MESSAGE_LEN];
	const char *pw_hist_file = krb5_config_get_string(context, NULL,
			"password_quality", "pw_history_file", NULL);
	int	reserved, i, j,
		min_time_diff = krb5_config_get_int_default(context, NULL, 0,
			"password_quality", "pwchange_timediff" , NULL);
	time_t	timestamp, cur_time;
	FILE	*pw_hist;

	val[0] = '\0';

	/* no check in password history wanted */
	if ( pw_hist_file == NULL ) return NULL;

	pw_hist = fopen(pw_hist_file, "r");
	if (!pw_hist)
	{
		strcpy(val, "password history database not available");
		goto PWREUSE_OUT;
	}
	fcntl(fileno(pw_hist), F_SETLKW);

	/* go through history database entry by entry */
	while ( fscanf(pw_hist, "%s %d %ld %s\n", account, &reserved,
		&timestamp, history) != EOF )
	{
		size_t len;

		/* is it the user's entry? */
		if ( strcmp(user, account) ) continue;

		cur_time = time(NULL);
		if ( (cur_time - timestamp) < (min_time_diff * 60) )
		{
			sprintf(val, "password already changed %d minutes ago "
				"(must be at least %d minutes)",
				(cur_time - timestamp) / 60, min_time_diff);
			goto PWREUSE_OUT;
		}

		len = strlen(history);
		/* check for reuse */
		for ( i = j = 0; i < len; i++ )
		{
			/* split the history at ',' */
			if ( (history[i] != ',') && (i != len - 1) ) continue;
			/* to get the the calculation right... */
			if ( i == len - 1 ) i++;

			/* save the current crypted password */
			strncpy(crypted, history + j, i - j);
			/* make it null-terminated */
			crypted[i - j] = '\0';
			/* is it reused? */
			if ( !check_pw_equal(crypted, password) )
			{
				strcpy(val, "it's a reused password");
				goto PWREUSE_OUT;
			}
			j = i + 1;
		}
		break;
	}

PWREUSE_OUT:
	if (pw_hist) fclose(pw_hist);
	return strlen(val) > 0 ? val : NULL;
}

/* generate a random salt */
void generate_salt(char * salt)
{
	char psalt[64], i, j = 0;

	for ( i = 'a'; i <= 'z'; i++ ) psalt[j++] = i;
	for ( i = 'A'; i <= 'Z'; i++ ) psalt[j++] = i;
	for ( i = '0'; i <= '9'; i++ ) psalt[j++] = i;
	psalt[j++] = '.';

	srand(time(NULL));
	for ( i = 0; i < 2; i++ ) salt[i] = psalt[rand() % sizeof(psalt)];
	salt[2] = '\0';
}

char * save_pw_history(krb5_context context, char * user, char * password)
{
	char	account[32], salt[3], history[(MAX_HIST + 1) * 14 + 1],
		*new_history = NULL, *pw_hist_file_tmp = NULL;
	static char val[MESSAGE_LEN];
	int	reserved = 0, found = 0,
		max_hist = krb5_config_get_int_default(context, NULL, 5,
			"password_quality", "pw_history", NULL);
	time_t	timestamp;
	FILE	*pw_hist = NULL, *pw_hist_tmp = NULL;
	const char *pw_hist_file = krb5_config_get_string(context, NULL,
			"password_quality", "pw_history_file", NULL);

	val[0] = '\0';

	/* no check in password history wanted */
	if ( pw_hist_file == NULL ) return NULL;

	if ( !( pw_hist_file_tmp = malloc(strlen(pw_hist_file) + 5) ) )
	{
		strcpy(val, "out of memory");
		goto SAVE_PWREUSE_OUT;
	}
	strcpy(pw_hist_file_tmp, pw_hist_file);
	strcat(pw_hist_file_tmp, ".tmp");

	/* another process just rewrites this file
	   wait until it has finished */
	while ( (pw_hist_tmp = fopen(pw_hist_file_tmp, "r")) )
	{
		fclose(pw_hist_tmp);
		if ( reserved++ > 1000 )
		{
			strcpy(val, "lock file exists, giving up");
			goto SAVE_PWREUSE_OUT;
		}
		usleep(100);
	}

	/* rename the original file */
	rename(pw_hist_file, pw_hist_file_tmp);
	if (! (pw_hist_tmp = fopen(pw_hist_file_tmp, "r")) )
	{
		strcpy(val, "password history database not available");
		goto SAVE_PWREUSE_OUT;
	}

	/* open the new file */
	if (! (pw_hist = fopen(pw_hist_file, "w")) )
	{
		rename(pw_hist_file_tmp, pw_hist_file);
		strcpy(val, "cannot modify password history database");
		goto SAVE_PWREUSE_OUT;
	}
	/* set an exclusive lock on this file, wait if it is already locked */
	fcntl(fileno(pw_hist), F_SETLKW);

	generate_salt(salt);

	/* go through history database entry by entry */
	while ( fscanf(pw_hist_tmp, "%s %d %ld %s\n", account, &reserved,
		&timestamp, history) != EOF )
	{
		new_history = history;
		/* is it the user's entry? */
		if ( !strcmp(user, account) )
		{
			found = 1;
			/* append new crypted password to history */
			strcat(history, ",");
			strcat(history, crypt(password, salt));

			/* too many old passwords? -> eliminate the oldest */
			new_history = (strlen(history) + 1) / 14 > max_hist ?
				history + strlen(history) + 1 - max_hist * 14 :
				history;
			timestamp = time(NULL);
			reserved = 0;
		}

		fprintf(pw_hist, "%s %d %ld %s\n", account,
			reserved, timestamp, new_history);
	}
	/* no entry yet */
	if (!found)
		fprintf(pw_hist, "%s %d %ld %s\n", user, 0,
			time(NULL), crypt(password, salt));

SAVE_PWREUSE_OUT:
	/* also performs fcntl(fileno(pw_hist), F_UNLCK)
	   regarding to man page */
	if ( pw_hist ) fclose(pw_hist);
	if ( pw_hist_tmp ) fclose(pw_hist_tmp);
	if ( pw_hist_file_tmp )
	{
		remove(pw_hist_file_tmp);
		free(pw_hist_file_tmp);
	}
	if ( strlen(val) > 0 )
		strcat(val, "!\nTurn to your system administrator");
	return strlen(val) > 0 ? val : NULL;
}

char * notify_socket(krb5_context context, char * user)
{
	const char *socket_name = krb5_config_get_string(context, NULL,
			"password_quality", "notify_socket", NULL);
	static char val[MESSAGE_LEN];
	char	data[MESSAGE_LEN];
	struct	sockaddr_un server;
	int	socket_fd = 0;

	val[0] = '\0';

	/* No notification wanted */
	if ( socket_name == NULL ) return NULL;

	if ( (socket_fd = socket(PF_UNIX, SOCK_STREAM, 0)) < 0 )
	{
		strcpy(val, "Could not create UNIX socket");
		goto NOTIFY_SOCKET_OUT;
	}

	server.sun_family = AF_UNIX;
	strcpy(server.sun_path, socket_name);

	if ( connect(socket_fd, (struct sockaddr *) &server,
		sizeof(struct sockaddr_un)) < 0 )
	{
		strcpy(val, "Could not connect to UNIX socket");
		goto NOTIFY_SOCKET_OUT;
	}

	sprintf(data, "%s\n", user);
	if ( write(socket_fd, data, strlen(data)) < 0 ) 
	{
		strcpy(val, "Could not write to UNIX socket");
		goto NOTIFY_SOCKET_OUT;
	}

	if ( read(socket_fd, data, sizeof(data)) < 0 ) 
	{
		strcpy(val, "Could not read from UNIX socket");
		goto NOTIFY_SOCKET_OUT;
	}

	if ( strncmp(data, "OK", 2) )
	{
		strcpy(val, "UNIX socket did not return a proper answer");
		goto NOTIFY_SOCKET_OUT;
	}

NOTIFY_SOCKET_OUT:
	if ( socket_fd > 0 ) close(socket_fd);
	if ( strlen(val) > 0 )
		strcat(val, "!\nTurn to your system administrator");
	return strlen(val) > 0 ? val : NULL;
}
